#include "sht.h"

#define HASHSIZE (1024*2)

/*
 *  Seed Hash Table
 *  -
 *  - 
 */



int nbGC(SEEDCODE word, int weight){
  int nb = 0;
  for(int i=0 ; i < weight ; i++){
    int code = (word >> 2 *(i)) & 0x3;
    if (code >= 2)
      nb++;
  }
  return nb;
}



sht::sht(string & seedPattern) : _seed_pattern(seedPattern), _sca_normal(seedPattern,false), _sca_reversed(_seed_pattern,true)  {
  /* (1) Hash table counters */
  _map_count    = map_type();
}


sht::sht(iodna &io, string & seedPattern) : _seed_pattern(seedPattern), _sca_normal(seedPattern,false), _sca_reversed(_seed_pattern,true) {
  
  /* (1) Hash table counters */
  _map_count    = map_type();
 
  /* (2) add a sequence */
  add(io);
}

sht::~sht() {
  _map_count.clear();
}

int sht::clear() {
  _map_count.clear();
  return 0;
}




int sht::add(iodna &io, SEEDCODE prefix , SEEDCODE mask ) {
  
  vector<seqdna> sequences = io.getSequences();
  
  for( int i  = 0 ; i < (int) sequences.size(); i++){
    
    _sca_normal.init(sequences[i]);
   
    while(_sca_normal.hasNext()){
      SEEDCODE code = _sca_normal.next();

      if ((code & mask) == prefix){ 
	
	map_type::iterator it = _map_count.find(code);
	if (it != _map_count.end()){
#ifdef MASKED
	  it->second.count_all++;
	  if (_sca_normal.isMasked())
	    it->second.count_masked++;
#else
	  it->second++; 
#endif
	  
	} else {
	 
#ifdef MASKED 
	  if (_sca_normal.isMasked()){
	    map_type::value_type p(code,Count(1,1));
	    _map_count.insert(p);
	  }else{
	    map_type::value_type p(code,Count(1,0));
	    _map_count.insert(p);
	  }
#else
	  map_type::value_type p(code,1);
	  _map_count.insert(p);
#endif

	}
      }
    }
    
    
    
#ifdef REVERSED
    
    _sca_reversed.init(sequences[i]);
    while(_sca_reversed.hasNext()){
      SEEDCODE code = _sca_reversed.next();
      
      if ((code & mask) == prefix){ 
	
	map_type::iterator it = _map_count.find(code);
	if (it != _map_count.end()){
#ifdef MASKED
	  it->second.count_all ++;
	  if (_sca_reversed.isMasked())
	    it->second.count_masked++;
#else
	  it->second++; 
#endif
	  
	} else {
	  
#ifdef MASKED 
	  if (_sca_reversed.isMasked()){
	    map_type::value_type p(code,Count(1,1));
	    _map_count.insert(p);
	  }else{
	    map_type::value_type p(code,Count(1,0));
	    _map_count.insert(p);
	  }
#else
	  map_type::value_type p(code,1);
	  _map_count.insert(p);
#endif

	  
	}
      }
    }
#endif
  }
  return 0;
}



std::ostream& sht::content(std::ostream& os, int feature_start) {
  
  for(map_type::iterator 
      it  = _map_count.begin();
      it != _map_count.end();
      it++){
    
#ifdef MASKED
    Count c =  it->second;
    SEEDCOUNT count          = c.count_all;
    SEEDCOUNT count_masked   = c.count_masked;
    SEEDCOUNT count_unmasked = count - count_masked;
#else
    SEEDCOUNT count = it->second;
#endif	
    os << "\t" << (it->first + feature_start + 1) << ":" << count;
  }
  return os;
}



std::ostream& sht::count(std::ostream& os) {

  cerr << "# count ... " << endl;  
  SEEDCOUNT total_count = 0 ;
  vector<SEEDCOUNT>         vcount  = vector<SEEDCOUNT>();
#ifdef MASKED
  vector<double>            vucount = vector<double>();
  vector<double>            vmcount = vector<double>();
#endif
#ifdef GCCOUNT
  vector< vector<SEEDCOUNT> > vcountGC = vector< vector<SEEDCOUNT> >(); /* compte les mots en fonction du nombre de GC */
#endif

  for(map_type::iterator 
	it  = _map_count.begin();
        it != _map_count.end();
      it++){

#ifdef MASKED
    Count c =  it->second;
    SEEDCOUNT count          = c.count_all;
    SEEDCOUNT count_masked   = c.count_masked;
    SEEDCOUNT count_unmasked = count - count_masked;
#else
    SEEDCOUNT count = it->second;
#endif	
    
    total_count += count;
    while(count >= vcount.size()){
      vcount.push_back(0);

#ifdef MASKED
      vmcount.push_back(0);	
      vucount.push_back(0);
#endif

#ifdef GCCOUNT
      vcountGC.push_back(vector<SEEDCOUNT>(1+getBitWeight()/2));
#endif

    }
    
#ifdef GCCOUNT
    int nbGCc = nbGC(it->first,getBitWeight()/2);
    vcountGC[count][nbGCc]++;
#endif

    vcount[count]++;
    
#ifdef MASKED	
    vmcount[count] += (double)count_masked;
    vucount[count] += (double)count_unmasked;
#endif    
    
  }//for



  // print first line = nb elem
  int nb_elem = 0 ;
  for(unsigned i=0; i < vcount.size(); i++) {
    if (vcount[i] > 0){
      nb_elem++;
    }
  }
  os << nb_elem << endl;
  
  
  // print second line = key frequency
  for(unsigned i=0; i < vcount.size(); i++) {
    if (vcount[i] > 0)
      os << (i) << "\t";
  }
  os << endl;

  
  // print third line = key ratio
  for(unsigned i=0; i < vcount.size(); i++) {
    if (vcount[i] > 0)
      os << (vcount[i]) << "\t";
  }
  os << endl;

#ifdef MASKED
  
  // print added line = masked sum
  for(unsigned i=0; i < vcount.size(); i++) {
    if (vcount[i] > 0)
      os << (vmcount[i]) << "\t";
  }
  os << endl;

  // print added line = unmasked sum
  for(unsigned i=0; i < vcount.size(); i++) {
    if (vcount[i] > 0)
      os << (vucount[i]) << "\t";
  }
  os << endl;

#endif

#ifdef GCCOUNT

  // print last lines = count 
  for(int c=0 ; c <= getBitWeight()/2   ; c++) {
    //os << "# count GC :" << c << endl;
    for(unsigned i=0; i < vcount.size(); i++) {
      if (vcount[i] > 0){
	os << (vcountGC[i][c])  << "\t";
      }
    }
    os << endl;
  }
 
#endif
  
  return os;
}
  



int sht::writeBinary(char * name) {
  cerr << "#" << endl;  
  fstream binary_file(name,ios::out|ios::binary|ios::app);
  for(map_type::iterator 
      it  = _map_count.begin();
      it != _map_count.end();
      it++){
    SEEDCODE  code  = it->first; 
    binary_file.write((char *)(&code),sizeof(SEEDCODE));
#ifdef MASKED	
    Count     count = it->second;
    binary_file.write((char *)(&count),sizeof(Count));
#else
    SEEDCOUNT count = it->second;
    binary_file.write((char *)(&count),sizeof(SEEDCOUNT));
#endif		
  }
  binary_file.close();
  return 0;
}
  
  

 

std::ostream& operator<<(std::ostream& os, sht & sht){
  cerr << "--- count ---" << endl;  
  vector<int> vcount = vector<int>();
  for(map_type::iterator 
      it  = sht._map_count.begin();
      it != sht._map_count.end();
      it++){
   
#ifdef MASKED	
    SEEDCOUNT count = it->second.count_all;
#else
    SEEDCOUNT count = it->second;
#endif	

    while(count >= vcount.size()){
      vcount.push_back(0);
    }    
    vcount[count]++;    
  }
  
 

  // print first line = key frequency
  for(unsigned i=0; i < vcount.size(); i++) {
    if (vcount[i] > 0)
      os << (i) << "\t";
  }
  os << endl;
  
  // print second line = key ratio
  for(unsigned i=0; i < vcount.size(); i++) {
    if (vcount[i] > 0)
      os << (double)(vcount[i]) << "\t";
  }
  os << endl;

  return os;
}

